/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.xml;

import org.opengion.fukurou.util.Closer ;
import org.opengion.fukurou.util.FileUtil ;

import java.io.PrintWriter ;
import java.io.BufferedWriter ;
import java.io.OutputStreamWriter ;
import java.io.FileOutputStream ;
import java.io.IOException ;
import java.io.File;
import java.io.StringReader ;
import java.io.FileNotFoundException ;
import java.io.UnsupportedEncodingException;
import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.ext.DefaultHandler2;
import org.xml.sax.InputSource ;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.ParserConfigurationException;


/**
 * JSP/XMLファイルを読み取って、OGNode/OGElement オブジェクトを取得する、ユーティリティークラスです。
 *
 * javax.xml.parsers および、org.w3c.dom の簡易処理を行います。
 * read で、トップレベルの OGNode を読み込み、write で、ファイルに書き出します。
 * 通常の W3C 系の オブジェクトを利用しないのは、属性の並び順を保障するためです。
 * ただし、属性のタブ、改行は失われます。
 * また、属性値に含まれるCR(復帰), LF(改行), TAB(タブ)は、 半角スペースに置き換えられます。
 * これは、SAXParser 側での XML の仕様の関係で、属性は、正規化されるためです。
 *
 * @og.rev 5.1.8.0 (2010/07/01) 新規作成
 *
 * @version  5.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK6.0,
 */
public class JspSaxParser extends DefaultHandler2 {
	public final static String CR = System.getProperty("line.separator");

	private String  attTab = "";		// tagBefore の一次TEMP

	private Stack<OGNode> stack = null;
	private OGNode  ele      = null;	// 現時点のエレメントノード
	private boolean inCDATA  = false;	// CDATA エレメントの中かどうかの判定
	private boolean inEntity = false;	// Entity の中かどうかの判定
	private File    file     = null;	// 処理実行中のファイル名

	/**
	 * XMLファイルを読み込み、OGDocument を返します。
	 *
	 * 内部的には、SAXParserFactory から、SAXParser を構築し、Property に、
	 * http://xml.org/sax/properties/lexical-handler を設定しています。
	 * コメントノードを処理するためです。
	 *
	 * @param aFile File
	 * @return OGDocument ファイルから読み取って構築した OGDocument
	 */
	public static OGDocument read( final File aFile ) {

		JspSaxParser sxp = new JspSaxParser();
		sxp.setFile( aFile );

		try {
			// SAXパーサーファクトリを生成
			SAXParserFactory spfactory = SAXParserFactory.newInstance();

			// SAXパーサーを生成
			SAXParser parser = spfactory.newSAXParser();

			parser.setProperty("http://xml.org/sax/properties/lexical-handler", sxp);	// LexicalHandler として
			// XMLファイルを指定されたハンドラーで処理します
			parser.parse( aFile, sxp );

		} catch (SAXNotRecognizedException ex) {
			String errMsg = "XMLReader は、認識されない機能またはプロパティー識別子を検出しました。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch (ParserConfigurationException ex) {
			String errMsg = "重大な構成エラーが発生しました。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch (SAXNotSupportedException ex) {
			String errMsg = "XMLReader は、要求された操作 (状態または値の設定) を実行できませんでした。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch (SAXException ex) {
			String errMsg = "SAX の一般的なエラーが発生しました。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch (IOException ex) {
			String errMsg = "ファイル読取時にエラーが発生しました。" 
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch( Exception ex) {
			String errMsg = "エラーが発生しました。" 
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		}

		return sxp.getDocument() ;
	}

	/**
	 * XML形式で表現された、文字列(String) から、OGDocument を構築します。
	 *
	 * 処理的には、#read( File ) と同じで、取り出す元が、文字列というだけです。
	 * XMLファイルからの読み込みと異なり、通常は、Element を表現した文字列が作成されますが、
	 * 返されるのは、OGDocument オブジェクトです。
	 *
	 * @param str String
	 * @return OGDocument ファイルから読み取って構築した OGDocument
	 */
	public static OGDocument string2Node( final String str ) {

		JspSaxParser sxp = new JspSaxParser();

		try {
			// SAXパーサーファクトリを生成
			SAXParserFactory spfactory = SAXParserFactory.newInstance();
			// SAXパーサーを生成
			SAXParser parser = spfactory.newSAXParser();

			parser.setProperty("http://xml.org/sax/properties/lexical-handler", sxp);	// LexicalHandler として

			// XMLファイルを指定されたデフォルトハンドラーで処理します
			InputSource source = new InputSource( new StringReader( str ) );
			parser.parse( source, sxp );

		} catch (SAXNotRecognizedException ex) {
			String errMsg = "XMLReader は、認識されない機能またはプロパティー識別子を検出しました。"
					+ CR + ex.getMessage();
			throw new RuntimeException( errMsg,ex );
		} catch (ParserConfigurationException ex) {
			String errMsg = "重大な構成エラーが発生しました。"
					+ CR + ex.getMessage();
			throw new RuntimeException( errMsg,ex );
		} catch (SAXException ex) {
			String errMsg = "SAX の一般的なエラーが発生しました。"
					+ CR + ex.getMessage();
			throw new RuntimeException( errMsg,ex );
		} catch (IOException ex) {
			String errMsg = "ストリームオブジェクト作成時にエラーが発生しました。" 
					+ CR + ex.getMessage();
			throw new RuntimeException( errMsg,ex );
		} catch( Exception ex) {
			String errMsg = "エラーが発生しました。" 
					+ CR + ex.getMessage();
			throw new RuntimeException( errMsg,ex );
		}

		return sxp.getDocument() ;
	}

	/**
	 * OGDocument を所定のファイルに、XML形式で書き出します。
	 *
	 * ここでは、UTF-8 文字コードでの書き出しです。
	 *
	 * @param aFile File      書き出すファイル
	 * @param node OGDocument 書き出す OGDocument
	 */
	public static void write( final File aFile, final OGDocument node ) {
		write( aFile,node,"UTF-8" );
	}

	/**
	 * OGDocument を所定のファイルに、XML形式で書き出します。
	 *
	 * @param aFile File      書き出すファイル
	 * @param node OGDocument 書き出す OGDocument
	 * @param encode String   書き出す文字エンコード(例：UTF-8）
	 */
	public static void write( final File aFile, final OGDocument node, final String encode ) {
		PrintWriter out = null;
		try {
			out = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream(aFile),encode )));
			out.println( "<?xml version=\"1.0\" encoding=\"" + encode + "\"?>" );
			out.println( node.toString() );
		} catch (FileNotFoundException ex) {
			String errMsg = "指定されたパス名で示されるファイルが存在しませんでした。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch (UnsupportedEncodingException ex) {
			String errMsg = "文字のエンコーディング(UTF-8)がサポートされていません。"
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		} catch( Exception ex) {
			String errMsg = "エラーが発生しました。" 
					+ CR + "\t" + ex.getMessage()
					+ CR + "\t" + aFile ;
			throw new RuntimeException( errMsg,ex );
		}
		finally {
			Closer.ioClose( out );
		}
	}

	/**
	 * ディレクトリの再帰処理でパース処理を行います。
	 *
	 * @param fromFile String 読み取りもとのファイル/フォルダ
	 * @param toFile String   書き込み先のファイル/フォルダ
	 */
	public static void copyDirectry( final File fromFile, final File toFile ) {
		// コピー元がファイルの場合はコピーして、終了する。
		if( fromFile.exists() && fromFile.isFile() ) {
			boolean isOK = false;
			String name = fromFile.getName();
			if( name.endsWith( ".jsp" ) || name.endsWith( ".xml" ) ) {
				try {
					OGDocument doc = JspSaxParser.read( fromFile );
					if( doc != null ) {
						JspSaxParser.write( toFile,doc );
						isOK = true;
					}
				}
				catch( Exception ex ) {
			//		ex.printStackTrace();
					System.out.println( ex.getMessage() );
					
				}
			}

			// JSPやXMLでない、パースエラー、書き出しエラーなど正常終了できなかった場合は、バイナリコピー
			if( !isOK ) {
				FileUtil.copy( fromFile,toFile,true );
			}
			return ;
		}

		// コピー先ディレクトリが存在しなければ、作成する
		if( !toFile.exists() ) {
			if( !toFile.mkdirs() ) {
				System.err.println( toFile + " の ディレクトリ作成に失敗しました。" );
				return ;
			}
		}

		// ディレクトリ内のファイルをすべて取得する
		File[] files = fromFile.listFiles();

		// ディレクトリ内のファイルに対しコピー処理を行う
		for( int i = 0; i<files.length; i++ ){
			copyDirectry( files[i], new File( toFile, files[i].getName()) );
		}
	}

	/**
	 * サンプルプログラムです。
	 *
	 * "Usage: JspSaxParser  <inFile|inDir> <outFile|outDir>"
	 *
	 * @param args String[]
	 */
	public static void main( final String[] args ) {
		if( args.length != 2 ) {
			System.out.println( "Usage: JspSaxParser  <inFile|inDir> <outFile|outDir>" );
		}

		File in   = new File( args[0] );
		File out  = new File( args[1] );

		JspSaxParser.copyDirectry( in,out );
	}

	/**
	 * 処理中のファイルオブジェクトを設定します。 
	 *
	 * これは、エラー、ワーニング時のファイル名を出力するために利用しています。
	 *
	 * @param file File 処理中のファイルオブジェクト 
	 */
	public void setFile( final File file ) {
		this.file = file;
	}

	// ********************************************************************************************** //
	// **                                                                                          ** //
	// ** ここから下は、DefaultHandler2 の実装になります。                                         ** //
	// **                                                                                          ** //
	// ********************************************************************************************** //

	/**
	 * 文書の開始通知を受け取ります。 
	 *
	 * インタフェース ContentHandler 内の startDocument
	 *
	 * @see org.xml.sax.helpers.DefaultHandler#startDocument()
	 * @see org.xml.sax.ContentHandler#startDocument()
	 */
	@Override
	public void startDocument() {
		stack	= new Stack<OGNode>();
		ele		= new OGDocument();
	}

	/**
	 * 要素の開始通知を受け取ります。 
	 *
	 * インタフェース ContentHandler 内の startElement
	 *
	 * @param uri String 名前空間 URI。要素が名前空間 URI を持たない場合、または名前空間処理が実行されない場合は null
	 * @param localName String 前置修飾子を含まないローカル名。名前空間処理が行われない場合は空文字列
	 * @param qName String 接頭辞を持つ修飾名。修飾名を使用できない場合は空文字列
	 * @param attributes Attributes 要素に付加された属性。属性が存在しない場合、空の Attributes オブジェクト 
	 *
	 * @see org.xml.sax.helpers.DefaultHandler#startElement(String,String,String,Attributes)
	 * @see org.xml.sax.ContentHandler#startElement(String,String,String,Attributes)
	 */
	@Override
	public void startElement( final String uri, final String localName, final String qName, final Attributes attributes ) {

		OGElement newEle = new OGElement( qName,attTab,attributes );
		ele.addNode( newEle );
		stack.push( ele );
		ele = newEle ;
	}

	/**
	 * 要素内の文字データの通知を受け取ります。 
	 *
	 * エンティティー内かどうかを判断する、inEntity フラグが true の間は、
	 * 何も処理しません。
	 *
	 * インタフェース ContentHandler 内の characters
	 *
	 * @param cbuf char[] 文字
	 * @param off int 文字配列内の開始位置
	 * @param len int 文字配列から使用される文字数
	 *
	 * @see org.xml.sax.helpers.DefaultHandler#characters(char[],int,int)
	 * @see org.xml.sax.ContentHandler#characters(char[],int,int)
	 */
	@Override
	public void characters( final char[] cbuf, final int off, final int len ) {
		if( inEntity ) { return ; }		// &lt; ⇒ < に変換されるので、エンティティ内では、なにも処理しない。

		String text = toText( cbuf,off,len );
		if( inCDATA ) {
			ele.addNode( text );
			return ;
		}

		OGNode node = new OGNode( text );
		ele.addNode( node );

		// '\r'（CR:復帰）+ '\n'（LF:改行）の可能性があるが、 '\n'（LF:改行）が、より後ろにあるので、これで判定。
		int lastIdx = text.lastIndexOf( '\n' );
		if( lastIdx >= 0 ) {
			attTab = text.substring( lastIdx+1 );	// 改行から、最後までの部分文字列
		}
		else {
			attTab = text;							// 改行がないので、すべて
		}
	}

	/**
	 * CDATA セクションの開始を報告します。
	 *
	 * CDATA セクションのコンテンツは、正規の characters イベントを介して報告されます。 
	 * このイベントは境界の報告だけに使用されます。
	 *
	 * インタフェース LexicalHandler 内の startCDATA
	 *
	 * @see org.xml.sax.helpers.DefaultHandler2#startCDATA()
	 * @see org.xml.sax.ext.LexicalHandler#startCDATA()
	 */
	@Override
	public void startCDATA() {
		OGNode node = new OGNode();
		node.setNodeType( OGNodeType.Cdata );

		ele.addNode( node );
		stack.push( ele );
		ele = node ;
		inCDATA = true;
	}

	/**
	 * CDATA セクションの終わりを報告します。
	 *
	 * インタフェース LexicalHandler 内の endCDATA
	 *
	 * @see org.xml.sax.helpers.DefaultHandler2#endCDATA()
	 * @see org.xml.sax.ext.LexicalHandler#endCDATA()
	 */
	@Override
	public void endCDATA() {
		ele = stack.pop();
		inCDATA = false;
	}

	/**
	 * DTD 宣言がある場合、その開始を報告します。 
	 *
	 * start/endDTD イベントは、ContentHandler の 
	 * start/endDocument イベント内の最初の startElement イベントの前に出現します。
	 *
	 * インタフェース LexicalHandler 内の startDTD
	 *
	 * @param name String 文書型名
	 * @param publicId String 宣言された外部 DTD サブセットの公開識別子。 宣言されていない場合は null
	 * @param systemId String 宣言された外部 DTD サブセットのシステム識別子。 宣言されていない場合は null。 
	 * 		  ドキュメントのベース URI に対しては解決されないことに 注意すること
	 * @see org.xml.sax.helpers.DefaultHandler2#startDTD( String , String , String )
	 * @see org.xml.sax.ext.LexicalHandler#startDTD( String , String , String )
	 */
	@Override
	public void startDTD( final String name, final String publicId, final String systemId ) {
		StringBuilder buf = new StringBuilder();
		buf.append( "<!DOCTYPE " ).append( name );
		if( publicId != null ) { buf.append( " PUBLIC \"" ).append( publicId ).append( "\"" ); }
		if( systemId != null ) { buf.append( "\"" ).append( systemId).append( "\"" ); }

		OGNode node = new OGNode( buf.toString() );
		node.setNodeType( OGNodeType.DTD );
		ele.addNode( node );
	}

	/**
	 * DTD 宣言の終わりを報告します。
	 *
	 * このメソッドは、DOCTYPE 宣言の終わりを報告するメソッドです。
	 * ここでは、何もしません。
	 *
	 * インタフェース LexicalHandler 内の endDTD
	 *
	 * @see org.xml.sax.helpers.DefaultHandler2#endDTD()
	 * @see org.xml.sax.ext.LexicalHandler#endDTD()
	 */
	@Override
	public void endDTD() {
		// ここでは何もしません。
	}

	/**
	 * 内部および外部の XML エンティティーの一部の開始を報告します。
	 *
	 * インタフェース LexicalHandler の記述:
	 *
	 * ※ ここでは、&amp;lt; などの文字列が、lt という名のエンティティーで
	 * 報告されるため、元の＆付きの文字列に復元しています。
	 * エンティティー内かどうかを判断する、inEntity フラグを true にセットします。
	 * inEntity=true の間は、#characters(char[],int,int) は、何も処理しません。
	 *
	 * @param name String  エンティティーの名前
	 * @see org.xml.sax.ext.LexicalHandler#startEntity(String)
	 */
	@Override
	public void startEntity( final String name ) {
		String text = "&" + name + ";" ;
		OGNode node = new OGNode( text );
		ele.addNode( node );
		inEntity = true;
	}

	/**
	 * エンティティーの終わりを報告します。 
	 *
	 * インタフェース LexicalHandler の記述:
	 *
	 * ※ ここでは、inEntity=false を設定するだけです。
	 *
	 * @param name String  エンティティーの名前
	 * @see org.xml.sax.ext.LexicalHandler#endEntity(String)
	 */
	@Override
	public void endEntity( final String name ) {
		inEntity = false;
	}

	/**
	 * 要素コンテンツに含まれる無視できる空白文字の通知を受け取ります。
	 *
	 * インタフェース ContentHandler 内の ignorableWhitespace
	 *
	 * @param cbuf char[] 空白文字
	 * @param off int  文字配列内の開始位置
	 * @param len int 文字配列から使用される文字数
	 *
	 * @see org.xml.sax.ContentHandler#ignorableWhitespace(char[],int,int)
	 */
	@Override
	public void ignorableWhitespace( final char[] cbuf, final int off, final int len ) {
		String text = toText( cbuf,off,len );
		OGNode node = new OGNode( text );
		ele.addNode( node );
	}

	/**
	 * 文書内の任意の位置にある XML コメントを報告します。
	 *
	 * インタフェース LexicalHandler の記述:
	 *
	 * @param cbuf char[] コメントに含まれる文字を格納する配列
	 * @param off int  配列内の開始位置
	 * @param len int 配列から読み取られる文字数
	 *
	 * @see org.xml.sax.helpers.DefaultHandler#characters(char[],int,int)
	 */
	@Override
	public void comment( final char[] cbuf, final int off, final int len ) {
		String text = toText( cbuf,off,len );
		OGNode node = new OGNode( text );
		node.setNodeType( OGNodeType.Comment );
		ele.addNode( node );
	}

	/**
	 * 要素の終了通知を受け取ります。
	 *
	 * @param uri String 名前空間 URI。要素が名前空間 URI を持たない場合、または名前空間処理が実行されない場合は null
	 * @param localName String 前置修飾子を含まないローカル名。名前空間処理が行われない場合は空文字列
	 * @param qName String 接頭辞を持つ修飾名。修飾名を使用できない場合は空文字列
	 *
	 * @see org.xml.sax.helpers.DefaultHandler#endElement(String,String,String)
	 * @see org.xml.sax.ContentHandler#endElement(String,String,String)
	 */
	@Override
	public void endElement( final String uri, final String localName, final String qName ) {
		ele = stack.pop();
	}

	/**
	 * パーサー警告の通知を受け取ります。
	 *
	 * インタフェース org.xml.sax.ErrorHandler 内の warning
	 *
	 * ここでは、パーサー警告の内容を標準エラーに表示します。
	 *
	 * @param ex SAXParseException 例外として符号化された警告情報
	 * @see org.xml.sax.ErrorHandler#warning(SAXParseException)
	 */
	@Override
	public void warning( final SAXParseException ex ) {
		String errMsg = ex.getMessage() + ":" + ex.getPublicId()
					+ CR + "\t" + file  + " (" + ex.getLineNumber() + ")";
		System.err.println( "WARNING:" + errMsg );
	}

	/**
	 * 文字配列から、文字列を作成します。（改行コードの統一）
	 *
	 * 処理的には、new String( cbuf,off,len ) ですが、XMLでリード
	 * されたファイルは、改行コードが、'\r'（CR:復帰）+ '\n'（LF:改行）ではなく、
	 * '\n'（LF:改行） のみに処理されます。（されるようです。規定不明）
	 * そこで、実行環境の改行コード（System.getProperty("line.separator")）と
	 * 置き換えます。
	 *
	 * @param cbuf char[] 文字を格納する配列
	 * @param off int  配列内の開始位置
	 * @param len int 配列から読み取られる文字数
	 * @return String 最終的な、Stringオブジェクト
	 */
	private String toText( final char[] cbuf, final int off, final int len ) {
		String text = new String( cbuf,off,len );
		return text.replaceAll( "\n", CR );
	}

	/**
	 * OGDocument を取得します。
	 *
	 * @return OGDocument 最終的な、OGNodeオブジェクト
	 */
	public OGDocument getDocument() {
		if( ele != null && ele.getNodeType() == OGNodeType.Document ) {
			return (OGDocument)ele ;
		}
		return null;
	}
}
